/*
 * ulpsm.c
 * linpsmѤΥץ
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>	/* open() */
#include <sys/stat.h>	/* open() */
#include <fcntl.h>	/* open(), O_RDONLY */

#include <assert.h>
#include "psm.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define O_BINARY 0

static struct opt {
	char *srcdev; 
	char *dstdev;
	int ifd, ofd;
} opt = {
	DEVICE_FILE,
	"/dev/null",
	-1, -1
};


static int psm_seek_frame(int fd, int frame)
{
	int code = lseek(fd, PSM_FRAME_SIZE * frame, SEEK_SET);
	if(code < 0)
		fprintf(stderr, "error: while seeking frame %d\n", frame);
	return code;
}

static int psm_seek_block(int fd, int block)
{
	int code = lseek(fd, PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK * block,
			 SEEK_SET);
	if(code < 0)
		fprintf(stderr, "error: while seeking block %d\n", block);
	return code;
}


static int psm_read_frame(int fd, void *buf, int frame)
{
	int tmp;

	tmp = psm_seek_frame(fd, frame);
	if(tmp < 0) return tmp;

	tmp = read(fd, buf, PSM_FRAME_SIZE);
	if(tmp != PSM_FRAME_SIZE)
		fprintf(stderr, "error: while reading frame %d\n", frame);
	return tmp != PSM_FRAME_SIZE;
}


static int psm_read_block(int fd, void *buf, int block)
{
	int tmp;

	tmp = psm_seek_block(fd, block);
	if(tmp < 0) return tmp;

	tmp = read(fd, buf, PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK);
	if(tmp != PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK)
		fprintf(stderr, "error: while reading block %d\n", block);
	return tmp != PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK;
}


static int psm_write_frame(int fd, void *buf, int frame)
{
	int tmp;

	tmp = psm_seek_frame(fd, frame);
	if(tmp < 0) return tmp;

	tmp = write(fd, buf, PSM_FRAME_SIZE);
	if(tmp != PSM_FRAME_SIZE)
		fprintf(stderr, "error: while writing frame %d\n", frame);
	return tmp != PSM_FRAME_SIZE;
}


static int psm_write_block(int fd, void *buf, int block)
{
	int tmp;

	tmp = psm_seek_block(fd, block);
	if(tmp < 0) return tmp;

	tmp = write(fd, buf, PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK);
	if(tmp != PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK)
		fprintf(stderr, "error: while writing block %d\n", block);
	return tmp != PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK;
}



/*----------------------------------------------------------------------
 */
static void psm_print_dir(struct opt *op)
{
	int i, j, tmp, next_block, data_bytes;
	struct dir_frame dir;
	struct title_frame title;
	char *err_msg;

	enum {
		dir_reserved = 'R',
		dir_free = 'f', dir_deleted = 'd',
		dir_start = 's', dir_continued = 'c', dir_end_chain = 'e',
		dir_error = 'E'
	} dir_stat;

	/*
	 * for each block...
	 */
	for(i = 1; i < 16; i++) {
		dir_stat = dir_error;
		err_msg = NULL;
		next_block = 0;
		data_bytes = 0;

		if(psm_read_frame(op->ifd, &dir, i)) {
			dir_stat = dir_error;
			err_msg = "failed to read directory frame";
			goto display;
		}

		/* type of directory entry */
		if(dir.dir_type == PSM_DIR_RESERVED)
			dir_stat = dir_reserved;

		else if((dir.dir_type & 0xf0) == PSM_DIR_AVAIL) {
			if((dir.dir_type & 0x0f) == PSM_DIR_FREE)
				dir_stat = dir_free;
			else
				dir_stat = dir_deleted;

		} else if((dir.dir_type & 0xf0) == PSM_DIR_INUSE) {
			if((dir.dir_type & 0x0f) == PSM_DIR_START)
				dir_stat = dir_start;
			else if((dir.dir_type & 0x0f) == PSM_DIR_CONTINUED)
				dir_stat = dir_continued;
			else if((dir.dir_type & 0x0f) == PSM_DIR_END_CHAIN)
				dir_stat = dir_end_chain;
		}

		/* check directory integrity */
		tmp = 0;
		for(j = 0; j < sizeof(dir); j++)
			tmp ^= *(j + (char*)&dir);
		if(tmp) {
			dir_stat = dir_error;
			err_msg = "directory csum error";
			goto display;
		}

		if(dir_stat == dir_start
		   && (dir.dir_sig[0] != 'B' || dir.dir_sig[1] != 'I')) {
			dir_stat = dir_error;
			err_msg = "signature error in directory frame";
			goto display;
		}

		/* next entry */
		tmp = (dir.dir_next[0] | dir.dir_next[1] << 8) + 1;
		next_block = tmp & 0xffff;

		/* size in bytes */
		data_bytes = dir.dir_size[0] << 16
			| dir.dir_size[1] << 8 | dir.dir_size[2];

		/*------------------------------------------------------------
		 * read title frame if start of data
		 */
		if(dir_stat != dir_start)
			goto display;

		if(psm_read_frame(op->ifd, &title, i * PSM_FRAMES_PER_BLOCK)) {
			dir_stat = dir_error;
			err_msg = "failed to read title frame";
			goto display;
		}

		if(title.title_sig[0] != 'S' || title.title_sig[1] != 'C') {
			dir_stat = dir_error;
			err_msg = "signature error in title frame";
			goto display;
		}

		/*------------------------------------------------------------
		 */
	  display:
		printf("%02d,%c,%02x,", i, dir_stat, dir.dir_type);
		if((dir.dir_type & 0x0f) >= PSM_DIR_START
		   && (dir.dir_type & 0x0f) <= PSM_DIR_END_CHAIN
		   && next_block != 0)
			printf("%d,", next_block);
		else 
			printf(",");
		printf("%04x,", data_bytes);
		if(dir_stat == dir_start || dir_stat == dir_deleted)
			printf("%.*s,", PSM_DIR_PRODUCT, dir.dir_product);
		else
			printf(",");

		if(dir_stat == dir_start
		   && (title.title_icons & 0xf0))
			printf("%d,%d,", title.title_icons & 0x0f,
			       title.title_icon_blocks);
		else
			printf("0,0,");

		if(dir_stat == dir_error)
			printf("%s\n", err_msg);
		else if(dir_stat == dir_start)
			printf("%.*s\n", PSM_TITLE_STRING,
			       title.title_string);
		else
			printf("\n");
	}
}


/*----------------------------------------------------------------------
 */
static int psm_copyall(struct opt *op)
{
	int i;
	char buf[PSM_FRAME_SIZE];

	fprintf(stderr, "source device: %s\n", op->srcdev);
	if(strcmp(op->dstdev, "-") == 0) {
		op->ofd = 1;	/* fileno(stdout) */
		fprintf(stderr, "destination device: standard input\n");
	} else {
		fprintf(stderr, "destination device: %s\n", op->dstdev);
		op->ofd = open(op->dstdev, O_WRONLY | O_CREAT | O_BINARY,
			       S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
		if(op->ofd < 0) {
			fprintf(stderr, "error: opening %s\n", op->dstdev);
			return 1;
		}
	}

	fprintf(stderr, "\nstatus: ");

	for(i = 0; i < PSM_BLOCKS_PER_CARD * PSM_FRAMES_PER_BLOCK; i++) {
		if(i % PSM_FRAMES_PER_BLOCK == 0)
			fprintf(stderr, "\n");

		fprintf(stderr, "  reading block %d frame %d  ",
			i / PSM_FRAMES_PER_BLOCK,
			i % PSM_FRAMES_PER_BLOCK);
		if(psm_read_frame(op->ifd, buf, i)) {
			perror("error: psm_copyall(read)");
			break;
		} else {
			fprintf(stderr, "\r");
		}

		fprintf(stderr, "  writing block %d frame %d  ",
			i / PSM_FRAMES_PER_BLOCK,
			i % PSM_FRAMES_PER_BLOCK);
		if(psm_write_frame(op->ofd, buf, i)) {
			perror("error: psm_copyall(write)");
			break;
		} else {
			fprintf(stderr, "\r");
		}
	}
	return 0;
}


/*----------------------------------------------------------------------
 */
static int psm_geticon(struct opt *op, int num)
{
	int i, j, tmp, nr_icon;
	struct dir_frame dir;
	struct title_frame title;
	struct {
		unsigned char r, g, b;
	} paldata[PSM_NR_PALETTES];

	/* some check */
	psm_read_frame(op->ifd, &dir, num);
	if(dir.dir_type != (PSM_DIR_INUSE | PSM_DIR_START)) {
		fprintf(stderr, "error: not starting block\n");
		return 1;
	}

	psm_read_frame(op->ifd, &title, num * PSM_FRAMES_PER_BLOCK);
	if((title.title_icons & 0xf0) == 0
	   || (title.title_icons & 0x0f) == 0) {
		fprintf(stderr, "error: no icon image\n");
		return 1;
	}

	/* get color palettes from title frame */
	nr_icon = title.title_icons & 0x0f;
	for(i = 0; i < PSM_NR_PALETTES; i++) {
		tmp = title.title_palette[i * 2 + 0]
			| (title.title_palette[i * 2 + 1] << 8);
		paldata[i].r = (tmp & 31) * 255 / 31;
		paldata[i].g = ((tmp >> 5) & 31) * 255 / 31;
		paldata[i].b = ((tmp >> 10) & 31) * 255 / 31;
	}

	/* output icons */
	for(i = 0; i < nr_icon; i++) {
		char buf[PSM_FRAME_SIZE];
		FILE *ofp;

		sprintf(buf, "psm%d-%d.xpm", num, i);
		ofp = fopen(buf, "w");

		/* write XPM header */
		sprintf(buf, "psm%d_%d", num, i);
		fprintf(ofp, "/* XPM */\n");
		fprintf(ofp, "static char *%s[] = {\n", buf);
		fprintf(ofp, "\"16 16 16 1\",\n");
		for(j = 0; j < PSM_NR_PALETTES; j++) {
			fprintf(ofp, "\"%c c #%02x%02x%02x\",\n",
				j + 'A',
				paldata[j].r, paldata[j].g, paldata[j].b);
		}

		/* write XPM pixel data */
		psm_read_frame(op->ifd, buf,
			       num * PSM_FRAMES_PER_BLOCK + i + 1);
		fprintf(ofp, "/* pixels */\n");
		for(j = 0; j < 16 * 16 / 2; j++) {
			if(j % 8 == 0) fprintf(ofp, "\"");
			fputc('A' + (buf[j] & 0x0f), ofp);
			fputc('A' + (buf[j] & 0xf0) / 16, ofp);
			if(j % 8 == 7) fprintf(ofp, "\",\n");
		}
		fprintf(ofp, "};\n");
		fclose(ofp);
	}
	return 0;
}


/*----------------------------------------------------------------------
 */
static int psm_get_block_chain(int fd, int start,
			       int chain[PSM_BLOCKS_PER_CARD])
{
	struct dir_frame dir;
	int next_block, nblk;

	nblk = 0;

	if(start < 1 || start >= PSM_BLOCKS_PER_CARD) {
		fprintf(stderr, "error: get_block_chain: starting block %d out of range\n", start);
		return -1;
	}
	psm_read_frame(fd, &dir, start);
	if(dir.dir_type != (PSM_DIR_INUSE | PSM_DIR_START)) {
		fprintf(stderr, "error: get_block_chain: block %d (type=0x%02x) not in use\n",
			start, dir.dir_type);
		return -2;
	}

	/* record starting block of data */
	chain[nblk++] = start;

	/* checks continued block */
	next_block = (dir.dir_next[0] | dir.dir_next[1] << 8) + 1;
	next_block &= 0xffff;
	if(next_block == 0)
		return nblk;

	/* rest of chains */
	for(;;) {
		if(next_block < 1 && next_block >= PSM_BLOCKS_PER_CARD) {
			fprintf(stderr, "error: get_block_chain: block %d out of range\n",
				next_block);
			return -3;
		}

		/* load next directory */
		psm_read_frame(fd, &dir, next_block);
		if(dir.dir_type != (PSM_DIR_INUSE | PSM_DIR_CONTINUED)
		   && dir.dir_type != (PSM_DIR_INUSE | PSM_DIR_END_CHAIN)) {
			fprintf(stderr,"error: cant link chain at block %d type=0x%02x\n",
				next_block, dir.dir_type);
			return -4;
		}

		/* record block linkage */
		chain[nblk++] = next_block;

		/* checks continued block */
		next_block = (dir.dir_next[0] | dir.dir_next[1] << 8) + 1;
		next_block &= 0xffff;
		if(next_block == 0
		   || dir.dir_type == (PSM_DIR_INUSE | PSM_DIR_END_CHAIN))
			break;
	}
	return nblk;
}


/*----------------------------------------------------------------------
 * ̤ѥ֥åΥꥹȤ
 * EEPROM顢ޤ٤ȤäƤʤ֥åͥŪ
 * ꥹȥåפ
 */
static int psm_get_free_blocks(int fd, int chain[PSM_BLOCKS_PER_CARD])
{
	int i, nblk;
	struct dir_frame dirs[PSM_BLOCKS_PER_CARD];

	nblk = 0;
	for(i = 1; i < PSM_BLOCKS_PER_CARD; i++)
		psm_read_frame(fd, &dirs[i], i);

	/* search blocks that is never used */
	for(i = 1; i < PSM_BLOCKS_PER_CARD; i++) {
		if(dirs[i].dir_type == (PSM_DIR_AVAIL | PSM_DIR_FREE))
			chain[nblk++] = i;
	}

	for(i = 1; i < PSM_BLOCKS_PER_CARD; i++) {
		if((dirs[i].dir_type & 0xf0) == PSM_DIR_AVAIL
		   && dirs[i].dir_type != (PSM_DIR_AVAIL | PSM_DIR_FREE))
			chain[nblk++] = i;
	}
#if 0
	for(i = 0; i < nblk; i++)
		fprintf(stderr, "free block %d = %d\n", i, chain[i]);
#endif

	return nblk;
}


/*----------------------------------------------------------------------
 */
static int psm_copy(struct opt *op, int num)
{
	int i, j, tmp;
	int used_blocks, free_blocks;
	int src_chain[PSM_BLOCKS_PER_CARD], dst_chain[PSM_BLOCKS_PER_CARD];
	struct dir_frame srcdir, dstdir;
	char blkbuf[PSM_FRAME_SIZE * PSM_FRAMES_PER_BLOCK];

	op->ofd = open(op->dstdev, O_RDWR | O_BINARY);
	if(op->ofd < 0) {
		fprintf(stderr, "error: can't open %s as dstdev\n", op->dstdev);
		return 1;
	}

	used_blocks = psm_get_block_chain(op->ifd, num, src_chain);
	free_blocks = psm_get_free_blocks(op->ofd, dst_chain);
	if(used_blocks > free_blocks) {
		fprintf(stderr, "error: can't copy %d data blocks into %d free blocks\n",
			used_blocks, free_blocks);
		return 2;
	}

	/* check if same product ID exists in destination device */
	psm_read_frame(op->ifd, &srcdir, num);
	for(i = 1; i < PSM_BLOCKS_PER_CARD; i++) {
		psm_read_frame(op->ofd, &dstdir, i);
		if(dstdir.dir_type == (PSM_DIR_INUSE | PSM_DIR_START)
		   && strncmp(srcdir.dir_product, dstdir.dir_product,
			      PSM_DIR_PRODUCT) == 0) {
			fprintf(stderr, "error: attempt to overwrite block (product ID is %s), please delete it first\n", srcdir.dir_product);
			return 1;
		}
	}

	/* copy directory and block */
	for(i = 0; i < used_blocks; i++) {
		fprintf(stderr, "src blk %d -> dst blk %d\n",
			src_chain[i], dst_chain[i]);
		psm_read_frame(op->ifd, &srcdir, src_chain[i]);
		tmp = ((i == used_blocks - 1) ? 0 : dst_chain[i+1]) - 1;
		srcdir.dir_next[0] = (char)tmp;
		srcdir.dir_next[1] = (char)(tmp >> 8);

		/* compute csum again */
		tmp = 0;
		for(j = 0; j < sizeof(srcdir) - 1; j++)
			tmp ^= *(j + (char*)&srcdir);
		srcdir.dir_xor = tmp;

		/* update block */
		psm_read_block(op->ifd, blkbuf, src_chain[i]);
		psm_write_block(op->ofd, blkbuf, dst_chain[i]);

		/* update directory */
		psm_write_frame(op->ofd, &srcdir, dst_chain[i]);
	}


	return 0;
}


/*----------------------------------------------------------------------
 * ꤵ줿ǡ֥åõ롣
 * ʣΥ֥åʤǡϡƬΥ֥åꤷʤФʤʤ
 * оݤȤʤǥХϡξsrcdev==dstdevǤ
 */
static int psm_delete(struct opt *op, int num)
{
	int i, j, tmp;
	int nblk, chain[PSM_BLOCKS_PER_CARD];
	struct dir_frame dir;

	close(op->ifd);
	/* TODO:̩ˤϡ񤭹Ѥ˺ƥץ󤹤ޤǤδ֡
	 * ¾ΥץPSMѤ뤳ȤǤ
	 */
	op->dstdev = op->srcdev;
	op->ofd = op->ifd = open(op->dstdev, O_RDWR | O_BINARY);
	if(op->ofd < 0) {
		fprintf(stderr, "error: can't open %s as dstdev\n", op->dstdev);
		return 1;
	}

	nblk = psm_get_block_chain(op->ofd, num, chain);
	if(nblk <= 0) return nblk;

	for(i = 0; i < nblk; i++) {
		/* printf("delete: marking block %d as deleted\n", chain[i]); */
		psm_read_frame(op->ofd, &dir, chain[i]);

		/* modify dir type and chksum*/
		tmp = 0;
		dir.dir_type = PSM_DIR_AVAIL | (dir.dir_type & 0x0f);
		for(j = 0; j < sizeof(dir) - 1; j++)
			tmp ^= *(j + (char*)&dir);
		dir.dir_xor = tmp;

		psm_write_frame(op->ofd, &dir, chain[i]);
	}
	return 0;
}


/*----------------------------------------------------------------------
 */
static int psm_init(struct opt *op)
{
	int i;
	union {
		struct dir_frame dir;
		char b[PSM_FRAME_SIZE];
	} u;

	close(op->ifd);
	op->dstdev = op->srcdev;
	op->ofd = open(op->dstdev, O_RDWR | O_BINARY);
	if(op->ofd < 0) {
		fprintf(stderr, "error: can't open %s as dstdev\n", op->dstdev);
		return 1;
	}

	memset(&u, 0, sizeof(u));
	u.b[0] = 'M';
	u.b[1] = 'C';
	u.b[127] = 'M' ^ 'C';
	psm_write_frame(op->ofd, &u, 0);

	memset(&u, 0, sizeof(u));
	u.dir.dir_type = PSM_DIR_AVAIL | PSM_DIR_FREE;
	u.dir.dir_next[0] = u.dir.dir_next[1] = 0xff;
	u.dir.dir_xor = PSM_DIR_AVAIL | PSM_DIR_FREE;
	for(i = 1; i < 16; i++)
		psm_write_frame(op->ofd, &u, i);

	return 0;
}


/*----------------------------------------------------------------------
 */
static int psm_check(struct opt *op)
{
	fprintf(stderr, "error: unimplemented yet, sorry _o_\n");
	return 1;
}

/*----------------------------------------------------------------------
 */
static void print_usage(void)
{
	fprintf(stderr, "ulpsm version " VERSION "\n");
	fprintf(stderr, "PlayStation(tm) memory card utility tool for Linux\n\n");
	fprintf(stderr, "Usage: \n");
	fprintf(stderr, 
"  ulpsm [-f SRCDEV] dir\n"
"  ulpsm [-f SRCDEV] ls\n"
"\tLists data block info in memory card file SRCDEV\n"
"  ulpsm [-f SRCDEV] copyall DSTDEV\n"
"  ulpsm [-f SRCDEV] dump DSTDEV\n"
"\tCopy all content to DSTDEV\n"
"  ulpsm [-f SRCDEV] geticon BLOCK\n"
"\tSave icon images from block BLOCK in XPM format\n"
"  ulpsm [-f SRCDEV] copy BLOCK DSTDEV\n"
"  ulpsm [-f SRCDEV] cp BLOCK DSTDEV\n"
"\tCopy data block to another memory card\n"
"  ulpsm [-f DSTDEV] delete BLOCK\n"
"  ulpsm [-f DSTDEV] rm BLOCK\n"
"\tDelete data block from memory card\n"
"  ulpsm [-f DSTDEV] format\n"
"  ulpsm [-f DSTDEV] newfs\n"
"\tErase ALL blocks\n"
);
}


/*----------------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
	int optidx = 1;
	char *missing_arg = "error: missing argument";

	/* check struct padding */
	assert(sizeof(struct dir_frame) == PSM_FRAME_SIZE);
	assert(sizeof(struct title_frame) == PSM_FRAME_SIZE);

	/* input device */
	if(argc > 2) {
		if(strcmp(argv[1], "-f") == 0) {
			opt.srcdev = argv[2];
			optidx += 2;
		}
	}

	/* no argument supplied */
	if(argc <= optidx) {
		print_usage();
		return 1;
	}

	/* open srcdev */
	opt.ifd = open(opt.srcdev, O_RDONLY | O_BINARY);
	if(opt.ifd < 0) {
		fprintf(stderr, "error: opening srcdev %s\n", opt.srcdev);
		return 1;
	}

	/* switch each action */
	if(strcmp(argv[optidx], "dir") == 0
	   || strcmp(argv[optidx], "ls") == 0) {
		psm_print_dir(&opt);

	} else if(strcmp(argv[optidx], "copyall") == 0
		  || strcmp(argv[optidx], "dump") == 0) {
		if(argc < optidx + 1) {
			fprintf(stderr, missing_arg);
			return 1;
		}
		opt.dstdev = argv[optidx + 1];
		psm_copyall(&opt);

	} else if(strcmp(argv[optidx], "geticon") == 0) {
		if(argc < optidx + 1) {
			fprintf(stderr, missing_arg);
			return 1;
		}
		psm_geticon(&opt, atoi(argv[optidx + 1]));

	} else if(strcmp(argv[optidx], "copy") == 0
		|| strcmp(argv[optidx], "cp") == 0) {
		if(argc < optidx + 3) {
			fprintf(stderr, missing_arg);
			return 1;
		}
		opt.dstdev = argv[optidx + 2];
		psm_copy(&opt, atoi(argv[optidx + 1]));

	} else if(strcmp(argv[optidx], "del") == 0
		  || strcmp(argv[optidx], "rm") == 0) {
		if(argc < optidx + 1) {
			fprintf(stderr, missing_arg);
			return 1;
		}
		psm_delete(&opt, atoi(argv[optidx + 1]));

	} else if(strcmp(argv[optidx], "format") == 0
		  || strcmp(argv[optidx], "newfs") == 0) {
		if(argc < optidx + 1) {
			fprintf(stderr, missing_arg);
			return 1;
		}
		psm_init(&opt);
	} else if(strcmp(argv[optidx], "check") == 0
		  || strcmp(argv[optidx], "fsck") == 0) {
		psm_check(&opt);
	} else {
		print_usage();
		return 1;
	}

	return 0;
}
